/*
	audio_macosx: audio output on MacOS X

	copyright ?-2006 by the mpg123 project - free software under the terms of the GPL 2
	see COPYING and AUTHORS files in distribution or http://mpg123.de
	initially written by tmkk
*/

#include <stdio.h>
#include <CoreServices/CoreServices.h>
#include <CoreAudio/CoreAudio.h>
#include <AudioUnit/AudioUnit.h>
#include <AudioToolbox/AudioToolbox.h>
#include <unistd.h>
#include <sys/time.h>

#include "mpg123.h"

OSStatus MyFileRenderProc(void *inRefCon, AudioUnitRenderActionFlags    *inActionFlags,
	const AudioTimeStamp *inTimeStamp, UInt32 inBusNumber,
	UInt32 inNumFrames, AudioBufferList *ioData);
OSStatus MyACComplexInputProc   (AudioConverterRef                        inAudioConverter,
	UInt32                               *ioNumberDataPackets,
	AudioBufferList                                   *ioData,
	AudioStreamPacketDescription  **outDataPacketDescription,
	void*                                          inUserData);

static AudioUnit outputUnit;
static AudioConverterRef converter;
static char *buffer1;
static char *buffer2;
static char *buffer1Top;
void *gSourceBuffer;
static int bufferIsFull;
static int buffer1Rest;
static int buffer2Rest;
static int lastBuffer;
static int playDone;
static int first;
static int channels;
static int softBufSize;

int copyBuffer(char *buffer, int bytes)
{
	int tmp;
	if(buffer1Rest > bytes) {
		memcpy(buffer1Top, buffer, bytes);
		buffer1Top = buffer1Top + bytes;
		buffer1Rest = buffer1Rest - bytes;
		return bytes;
	}
	memcpy(buffer1Top, buffer, buffer1Rest);
	buffer1Top = buffer1Top + buffer1Rest;
	tmp = buffer1Rest;
	buffer1Rest = 0;
	bufferIsFull = 1;
	if(first) {
		AudioOutputUnitStart(outputUnit);
		first = 0;
	}
	return tmp;
}

OSStatus MyFileRenderProc(void *inRefCon, AudioUnitRenderActionFlags    *inActionFlags,
                        const AudioTimeStamp *inTimeStamp, UInt32 inBusNumber,
                        UInt32 inNumFrames, AudioBufferList *ioData)
{
	OSStatus err= noErr;
	void *inInputDataProcUserData=NULL;
	AudioStreamPacketDescription* outPacketDescription =NULL;
	
	err = AudioConverterFillComplexBuffer(converter, MyACComplexInputProc ,inInputDataProcUserData , &inNumFrames, ioData, outPacketDescription);
	
	return err;
}

OSStatus MyACComplexInputProc   (AudioConverterRef                        inAudioConverter,
                                 UInt32                               *ioNumberDataPackets,
                                 AudioBufferList                                   *ioData,
                                 AudioStreamPacketDescription  **outDataPacketDescription,
                                 void*                                          inUserData)
{
	OSStatus    err = noErr;
	UInt32  bytesCopied = 0;
	int i;
	
	ioData->mBuffers[0].mData = NULL;      
	ioData->mBuffers[0].mDataByteSize = 0;
	
	if(!buffer2Rest && lastBuffer && !bufferIsFull) {
		playDone = 1;
		return noErr;
	}
	
	if(!buffer2Rest) {
		while(!bufferIsFull) usleep(10000);
		memcpy(buffer2, buffer1, softBufSize);
		buffer1Rest = softBufSize;
		buffer1Top = buffer1;
		bufferIsFull = 0;
		buffer2Rest = softBufSize;
	}
	
	if (*ioNumberDataPackets > buffer2Rest / channels / 2) 
		*ioNumberDataPackets = buffer2Rest / channels / 2;
	
	if (*ioNumberDataPackets)
	{
		if (gSourceBuffer != NULL) {
			free(gSourceBuffer);
			gSourceBuffer = NULL;
		}
		
		bytesCopied = *ioNumberDataPackets * 4;  
		gSourceBuffer = (void *) calloc (1, bytesCopied);
		
		if(channels == 1) {
			for(i=0;i<*ioNumberDataPackets;i++) {
				*((short *)(gSourceBuffer)+i*2) = *((short *)(buffer2 + softBufSize - buffer2Rest) + i);
				*((short *)(gSourceBuffer)+i*2+1) = *((short *)(buffer2 + softBufSize - buffer2Rest) + i);
			}
			buffer2Rest = buffer2Rest - bytesCopied/2;
		}
		else {
			memcpy(gSourceBuffer, buffer2 + softBufSize - buffer2Rest, bytesCopied);
			buffer2Rest = buffer2Rest - bytesCopied;
		}
		
		ioData->mBuffers[0].mData = gSourceBuffer;
		ioData->mBuffers[0].mDataByteSize = bytesCopied;
	}  
	else
	{
		ioData->mBuffers[0].mData = NULL;      
		ioData->mBuffers[0].mDataByteSize = 0;
		err = noErr;
	}
	
	return err;
}

int audio_open(struct audio_info_struct *ai)
{
	OSErr err;
	AudioStreamBasicDescription inFormat;
	AudioStreamBasicDescription outFormat;
	AURenderCallbackStruct  renderCallback;
	Boolean outWritable;
	UInt32		count ;
	ComponentDescription desc;
	Component comp;
	
	/* open default output nuit */
	desc.componentType = kAudioUnitType_Output; 
	desc.componentSubType = kAudioUnitSubType_DefaultOutput;
	desc.componentManufacturer = kAudioUnitManufacturer_Apple;
	desc.componentFlags = 0;
	desc.componentFlagsMask = 0;
	comp = FindNextComponent(NULL, &desc);
	if (comp == NULL) {
		fprintf(stderr, "can't find default output unit\n");
		return -1;
	}
	
	err = OpenAComponent(comp, &outputUnit);
	if (err)  {
		fprintf(stderr, "can't open default output unit\n");
		return -1;
	}
	
	err = AudioUnitInitialize(outputUnit);
	if(err != noErr) {
		fprintf(stderr, "AudioUnitInitialize failed.\n");
		return -1;
	}
	
	err = AudioUnitGetPropertyInfo(outputUnit, kAudioUnitProperty_StreamFormat, kAudioUnitScope_Output, 0, &count, &outWritable);
	err = AudioUnitGetProperty (outputUnit, kAudioUnitProperty_StreamFormat, kAudioUnitScope_Output, 0, &outFormat, &count);
	if(err != noErr) {
		fprintf(stderr, "AudioUnitGetProperty failed.\n");
		return -1;
	}
	AudioUnitSetProperty (outputUnit, kAudioUnitProperty_StreamFormat, kAudioUnitScope_Input, 0, &outFormat, count);
	if(err != noErr) {
		fprintf(stderr, "AudioUnitSetProperty failed.\n");
		return -1;
	}
	channels = ai->channels;
	int rate = ai->rate;
	if(rate == -1) rate = 44100;
	inFormat.mSampleRate = rate;
	inFormat.mBytesPerPacket = 4;
	inFormat.mFramesPerPacket = 1;
	inFormat.mBytesPerFrame = 4;
	inFormat.mChannelsPerFrame = 2;
	inFormat.mBitsPerChannel = 16;
	inFormat.mFormatID = kAudioFormatLinearPCM;
#ifdef _BIG_ENDIAN
	inFormat.mFormatFlags = kLinearPCMFormatFlagIsSignedInteger | kLinearPCMFormatFlagIsPacked | kLinearPCMFormatFlagIsBigEndian;
#else
	inFormat.mFormatFlags = kLinearPCMFormatFlagIsSignedInteger | kLinearPCMFormatFlagIsPacked;
#endif
	err = AudioConverterNew(&inFormat, &outFormat, &converter);
	if(err != noErr) {
		fprintf(stderr, "AudioConverterNew failed.\n");
		return -1;
	}
	
	memset(&renderCallback, 0, sizeof(AURenderCallbackStruct));
	renderCallback.inputProc = MyFileRenderProc;
	renderCallback.inputProcRefCon =0;
	err = AudioUnitSetProperty (outputUnit, kAudioUnitProperty_SetRenderCallback, kAudioUnitScope_Input, 0, &renderCallback, sizeof(AURenderCallbackStruct));
	if(err != noErr) {
		fprintf(stderr, "AudioUnitSetProperty failed.\n");
		return -1;
	}
	if(channels != -1) {
		softBufSize = 32768 * channels;
		if(buffer1Top == NULL) {
			if((buffer1 = (char *)malloc(softBufSize)) == NULL) {
				fprintf(stderr,"malloc error\n");
				return -1;
			}
			if((buffer2 = (char *)malloc(softBufSize)) == NULL) {
				fprintf(stderr,"malloc error\n");
				return -1;
			}
			buffer1Top = buffer1;
		}
	}
	
	bufferIsFull = 0;
	buffer1Rest = softBufSize;
	buffer2Rest = 0;
	lastBuffer = 0;
	playDone = 0;
	first = 1;
	
	return 0;
}

int audio_get_formats(struct audio_info_struct *ai)
{
	return AUDIO_FORMAT_SIGNED_16;
}

int audio_play_samples(struct audio_info_struct *ai,unsigned char *buf,int len)
{
	int copied,size=len;
	
	while(len) {
		while(bufferIsFull) usleep(10000);
		copied = copyBuffer((char *)buf,len);
		len = len - copied;
		buf = buf + copied;
	}
	return size;
}

void audio_queueflush(struct audio_info_struct *ai)
{
}

int audio_close(struct audio_info_struct *ai)
{
	if(!first) {
		memset(buffer1Top, 0, buffer1Rest);
		bufferIsFull = 1;
		lastBuffer = 1;
	}
	while(!playDone && !first) {
		usleep(10000);
	}
	AudioConverterDispose(converter);
	AudioOutputUnitStop(outputUnit);
	AudioUnitUninitialize (outputUnit);
	CloseComponent(outputUnit);

	
	return 0;
}